
:- [txf_strategy].

:- import str_cat/3 from string.
:- import length/2, member/2, append/3 from basics.
:- import conset/2, conget/2 from gensym.
:- dynamic proof_oblig/1.
:- dynamic unfold_prohibited/2.
:- dynamic no_unfold/1.
:- dynamic no_deduce/1.
:- dynamic proof_stack/2.
:- dynamic proved/2.

:- dynamic deduce_time/2.

%---------------------------------------------
%  Setting the Constants/Flags


fold_cond_max(2).      % Maximum length of conditions for 
                       % conditional folding.

equiv_cond_max(3).     % Maximum length of conditions for 
                       % conditional equivalence.

%--------------------------------------------------------------------

prove_all_pairs([]).

prove_all_pairs([Preds|Rest]) :-
	Preds = (Pred1, Pred2),
	prove(Pred1, Pred2),
	prove_all_pairs(Rest).

%-----------------------------------------------------------------
% proof obligation : Pred1 => Pred2


prove(Pred1, Pred2) :-
	proof_stack(Pred1, Pred2), !, fail.

prove(Pred1, Pred2) :-
	proved(Pred1, Pred2).

prove(Pred1, Pred2) :-
	assert(proof_stack(Pred1, Pred2)),
	save_pgm_state(Pgm, Markers), !,
	( txf_prove(Pred1, Pred2)  
	      -> restore_pgm_state(Pgm, Markers),
	         retract(proof_stack(Pred1, Pred2)),
	         myassert(proved(Pred1, Pred2))
	      ;  write('NO PROOF CONSTRUCTED: '), 
		 write(Pred1), writeln(' => '),
		 writeln(Pred2), fail
        ).

txf_prove(Pred1, Pred2) :-
	( not_axiom_chk(Pred1, Pred2) 
		-> transform(Pred1, Pred2), !,
		   txf_prove(Pred1, Pred2)
		;  true
        ).


%----------------------------------------------------------------------

% prove(Pred1, Pred2, Lo, Hi) is true if
%     1. Pred1 <=> Pred2
%     2. Lo \leq rank(Pred1) - rank(Pred2) \leq Hi


prove(Pred1, Pred2, Lo, Hi) :-
	assert(proof_stack(Pred1, Pred2)),
	txf_prove(Pred1, Pred2, Lo, Hi),
	retract(proof_stack(Pred1, Pred2)).


txf_prove(Pred1, Pred2, Lo, Hi) :-
	( axiom(Pred1, Pred2) 
		-> compute_proof_bounds(Pred1, Pred2, Lo, Hi)
		;  transform(Pred1, Pred2), !,
		   txf_prove(Pred1, Pred2, Lo, Hi)
        ).


%------------------------------------------------------------


transform(Pred1, Pred2) :- 
	(unfold_pred(Pred1) ; unfold_pred(Pred2)),
	conget(unf_cnt, T), T1 is T + 1, conset(unf_cnt, T1).

/*
transform(Pred1, Pred2) :-
	(fold_pred(Pred1) ; fold_pred(Pred2)),
	conget(txfcnt, T), T1 is T + 1, conset(txfcnt, T1).
*/

transform(Pred1, Pred2) :-
	cond_axiom(Pred1, Pred2).


unfold_pred(Pred) :-
	new_pred(Pred), \+ no_unfold(Pred), 
	find_clauses(Pred, CList),
	( algorithmic_transform(CList)
		-> %true
		   write_pred(Pred)
		;  assert(no_unfold(Pred)), fail
        ).


fold_pred(Pred) :-
	new_pred(Pred), \+ no_deduce(Pred), 
	find_clauses(Pred, CList),
	( deductive_transform(CList)
		-> true
		;  assert(no_deduce(Pred)), fail
        ).


algorithmic_transform(CList):-
	pick_clause(CList, C),
	select(C, Num), singleStepUnf(C, Num, CL),
	nl,writeln('UNFOLD'), write(C), write(' to '), writeln(CL),nl.

deductive_transform(CList) :- 
	cleanup_clauses(CList, CList1),
	fold(CList1, _).   % conditional folding involves repl.
			   % it is diabled for this one-way prover

%-------------------------------------------------------------------
% A predicate is deemed new (and hence suitable for transformation) if
% it has been marked by user as a proof obligation predicate or it is
% a newly introduced predicate during transformation.

new_pred(P/N) :- proof_oblig(P/N).

new_pred(P/_) :- atom_chars(P, PList), PList = [n,e,w|_].

%---------------------------------------------------------------

%pick_clause(CList, C) :- nl, writeln(CList), nl, read(C).

pick_clause(CList, C) :- member(C, CList).

write_pred(Pred1) :-
	write_star(32),
	Pred1 = P/N, 
	write_all(P, N),
	write_star(32).

%--------------------------------------------------

save_pgm_state(Pgm, Markers) :-
	save_current_clauses(Pgm),
	copy_defn_clauses,
	save_current_markers(Markers).

restore_pgm_state(Pgm, Markers) :-
	retractall(clauseI(_, _, _)),
	copy_old_clauses(Pgm),
	check_all_defs,  % any definitions introduced by the spawned
			 % proof is also asserted as current clause.
	retractall(no_unfold(_)),
	retractall(no_deduce(_)),
	copy_old_markers(Markers).

save_current_clauses(Pgm) :-
	findall(C, clauseI(C, _, _), CList),
	return_clauses(CList, Pgm),
	retractall(clauseI(_, _, _)).

return_clauses([], []).
return_clauses([C|Rest], [(C,B,A)|RestX]) :-
	clauseI(C, B, A), return_clauses(Rest, RestX).
	



save_current_markers((UList, FList)) :-
	findall(X, no_unfold(X), UList),
	findall(Y, no_deduce(Y), FList),
	retractall(no_unfold(_)),
	retractall(no_deduce(_)).


copy_defn_clauses :-
	clause0(X, Y, Z), assert(clauseI(X, Y, Z)), fail.
copy_defn_clauses.

copy_old_clauses([]).
copy_old_clauses([(X,Y,Z) | Rest]) :-
	assert(clauseI(X, Y, Z)), copy_old_clauses(Rest).

copy_old_markers((UList, FList)) :-
	copy_unf_markers(UList),
	copy_ded_markers(FList).


copy_unf_markers([]).
copy_unf_markers([X|Rest]) :-
	assert(no_unfold(X)), copy_unf_markers(Rest).

copy_ded_markers([]).
copy_ded_markers([X|Rest]) :-
	assert(no_deduce(X)), copy_ded_markers(Rest).

% after a spawned proof obligation returns successfully, we 
% take care to introduce any definitions "introduced" by the spawned
% proof obligation as clauses in the "current" program.
% Note that this is not required for correctness of the code
% (since all definitions are kept track by clause0/3 anyway).
% However, this is required for the "current" program to have
% have a clean semantics before and after the spawned oblig. returns.
		
check_all_defs :-
	defn_pred(P/N, DefList),
	length(Args, N), Hd =.. [P|Args],
	( clauseI(_, (Hd :- _), _) 
		-> true
		;  re_assert_def(DefList)
        ), fail.

check_all_defs.

re_assert_def([]).
re_assert_def([D|Rest]) :-
	clause0(D, B, A),
	assert(clauseI(D, B, A)),
	re_assert_def(Rest).



		